﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2023 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <math.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"
#include "String.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                    // Zeitphase:
{   TIMEPHASE_STD,              // · Normalzeit
    TIMEPHASE_DST,              // · Sommerzeit
    TIMEPHASE_UNKNOWN           // · Unbekannt
} TimePhase;

typedef enum                    // Sonnenereignisse:
{   SOLAREVENT_CIVIL_DAWN,      // · Bürgerliche Morgendämmerung
    SOLAREVENT_SUNRISE,         // · Sonnenaufgang
    SOLAREVENT_SUNSET,          // · Sonnenuntergang
    SOLAREVENT_CIVIL_DUSK       // · Bürgerliche Abenddämmerung
} SolarEvent;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool  updateSolarEvent = false;
static Time  civilDawnTime    = TIME_INVALID;
static Time  sunriseTime      = TIME_INVALID;
static Time  sunsetTime       = TIME_INVALID;
static Time  civilDuskTime    = TIME_INVALID;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen   //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  TimeShiftParameter(): Holt die für die Lokalisierung der Zeit notwendigen Para-     //
//  --------------------  meter.                                                        //
//                                                                                      //
//  Eingabe: · timeZone: Nimmt die aktuelle Zeitzone auf: -11 .. +12.                   //
//           · autoDst:  Erhält 'true', wenn bei Umrechnungen die Sommerzeitumstellung	//
//                       berücksichtigt werden soll.                                    //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
TimeShiftParameter(SInt08* timeZone, bool* autoDst)
{
    SInt08 storedTimeZone;
    STORING_Load(STO_TimeZone, &storedTimeZone);
    DSTMode dstMode;
    STORING_Load(STO_DstMode, &dstMode);
    SInt08 tmpTimeZone = 0;
    bool   tmpAutoDst  = false;
    switch (dstMode)
    {   case DSTMODE_OFF:
            tmpTimeZone = storedTimeZone;
            tmpAutoDst  = false;
            break;
        case DSTMODE_ON:
            tmpTimeZone = storedTimeZone+1;
            tmpAutoDst  = false;
            break;
        case DSTMODE_AUTO:
            tmpTimeZone = storedTimeZone;
            tmpAutoDst  = true;
            break;
        case DSTMODE_DCF:
            if (BOARD_RadioTimeZone() >= -11 && BOARD_RadioTimeZone() <= 12)
            {   tmpTimeZone = storedTimeZone+BOARD_RadioTimeZone()-1;
                tmpAutoDst  = false;
            }
            else
            {   tmpTimeZone = storedTimeZone;
                tmpAutoDst  = true;
            }
            break;
        default:
            ;
    }
    *timeZone = tmpTimeZone;
    *autoDst  = tmpAutoDst;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ReadLocalTime(): Die Uhrzeit wird vom Board-Modul in UTC zur Verfügung gestellt.    //
//  ---------------  Diese Funktion liest die UTC-Zeit aus und versetzt sie unter Be-   //
//  rücksichtigung der Sommerzeit-Betriebsart in die lokale Zeitzone. Ist die Systemuhr //
//  nicht gestellt, ist die zurückgelieferte Datumzeit ungültig.                        //
//                                                                                      //
//  Eingabe: · dateTime: Datumzeit, die die lokale Zeit und das Datum aufnimmt.         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ReadLocalTime(DateTime* dateTime)
{
    BOARD_ReadUTC(dateTime);
    if (!DATETIME_IsValid(dateTime)) return;
    SInt08 timeZone;
    bool   autoDst;
    TimeShiftParameter(&timeZone, &autoDst);
    DATETIME_Localize(dateTime, timeZone, autoDst);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  WriteLocalTime(): Umkehrfunktion zu ReadLocalTime(). Die in <dateTime> übergebene   //
//  ----------------  Zeit wird unter Berücksichtigung der eingestellten Zeitzone und   //
//  der Sommerzeit-Betriebsart in UTC umgerechnet und danach ins Boardmodul übertragen. //
//  Ist die in <dateTime> übergebene Zeit ungültig, bleibt die Funktion ohne Wirkung.   //
//  Hinweis: Bei automatischer Sommerzeitumstellung ist eine eindeutige Rückrechnung in //
//           UTC nicht immer möglich. Siehe dazu DATETIME_Universalize().               //
//                                                                                      //
//  Eingabe: · dateTime: Datumzeit, die die lokale Zeit und das Datum bereitstellt.     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
WriteLocalTime(DateTime* dateTime)
{
    if (!DATETIME_IsValid(dateTime)) return;
    SInt08 timeZone;
    bool   autoDst;
    TimeShiftParameter(&timeZone, &autoDst);
    DATETIME_Universalize(dateTime, timeZone, autoDst);
    BOARD_WriteUTC(dateTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CalcSolarTime(): Ermittelt den Zeitpunkt des Sonnen-Ereignisses in <event> an der   //
//  ---------------  durch <latitude> (geographische Breite) und <longitude> (geogra-   //
//  phische Länge) gegebenen Standort. Die Breite ist begrenzt auf ±65°, was in etwa    //
//  dem Bereich zwischen dem nördlichen und südlichen Polarkreis entspricht. Zur Be-    //
//  rechnung wird der Monat und Tag aus <date> sowie die lokale Zeitzone herangezogen.  //
//  Der ermittelte Zeitpunkt umfasst Stunden und Minuten als Ortszeit (Sekunden sind    //
//  immer 0). Konnte der Zeitpunkt nicht ermittelt ist die zurückgegebene Zeit un-      //
//  gültig.                                                                             //
//                                                                                      //
//  Eingabe: · solarTime: Sonnen-Ereignis gemäß Aufzählungstyp 'SolarEvent'.            //
//           · latitude:  Geographische Breite des Ortes: -6500 .. +6500 [°÷100].       //
//           · longitude: Geographische Länge  des Ortes: -17999 .. +18000 [°÷100].     //
//           · date :     Datum des zu ermittelnden Sonnenstandes (nur Monat und Tag).  //
//  Ausgabe: Zeitpunkt des Ereignisses, Sekunde ist immer 0.                            //
//////////////////////////////////////////////////////////////////////////////////////////
static Time
CalcSolarTime(SolarEvent event, SInt16 latitude, SInt16 longitude, const Date* date)
{
    Time time = TIME_INVALID;
    if (latitude  <  -6500 || latitude  >  6500) return time;
    if (longitude < -17999 || longitude > 18000) return time;
    int yearDay = DATE_YearDay(date);    // Tagnummer, beginnend mit 1 = 1. Januar
    // Deklination der Sonne (Breitengrad über dem die Sonne im Zenit steht)
    float declination = 0.4095*sin(0.016906*(yearDay-80.086));  // [rad]
    // Zeit, die die Sonne von Mittag bis zur Horizonthöhe <h> braucht
    float h;
    if (event == SOLAREVENT_SUNRISE || event == SOLAREVENT_SUNSET)
    {   // Sonnen-Horizont (60 Bogenminuten)
        h = -0.0145; // [rad]
    }
    else
    {   // Bürgerliche Dämmerung (-6°)
        h = -0.1047; // [rad]
    }
    const float pi = M_PI;
    float latInRad = (float)latitude/100 / 57.29578; // Breitengrad in [rad]
    float duration = 12*acos((sin(h) - sin(latInRad)*sin(declination )) /
    (cos(latInRad)*cos(declination )))/pi;
    // Wahre Sonnenzeit
    float apparentSolarTime = 12;
    switch (event)
    {   case SOLAREVENT_CIVIL_DAWN: apparentSolarTime -= duration; break;
        case SOLAREVENT_SUNRISE:    apparentSolarTime -= duration; break;
        case SOLAREVENT_SUNSET:     apparentSolarTime += duration; break;
        case SOLAREVENT_CIVIL_DUSK: apparentSolarTime += duration; break;
        default:                    return time;
    }
    // Zeitgleichung (Differenz zwischen wahrer und mittlerer Sonnenzeit)
    float equationOfTime = -0.1710*sin(0.03370 * yearDay + 0.465) -
    0.1299*sin(0.01787 * yearDay - 0.168); // [h]
    // Mittlere Sonnenzeit
    float meanSolarTime = apparentSolarTime - equationOfTime;
    // Mittlere Ortszeit (mittlere Sonnenzeit – geographische Länge÷15)
    float localMeanTime = meanSolarTime - (float)longitude/100/15; // [h]
    // Umwandlung in Stunden und Minuten
    time.hour   = (UInt08)localMeanTime;
    time.minute = (UInt08)((((UInt16)(localMeanTime*100+0.5)%100*6)+5)/10);
    time.second = 0;
    // Überführung in Ortszeit (MEZ/MESZ)
    SInt08 timeZone;
    bool   autoDst;
    TimeShiftParameter(&timeZone, &autoDst);
    DateTime dateTime = DATETIME(*date, time);
    DATETIME_Localize(&dateTime, timeZone, autoDst);
    return dateTime.time;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CurrentTimePhase(): Ermittelt die aktuell vorherrschende Zeitphase (Sommerzeit/     //
//  ------------------  Normalzeit). Die in den Einstellungen vorgenommene Sommerzeit-  //
//  Betriebsart wird berücksichtigt.                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktuelle Zeitphase als 'TimePhase'.                                        //
//////////////////////////////////////////////////////////////////////////////////////////
static TimePhase
CurrentTimePhase(void)
{
    DateTime  dateTime;
    BOARD_ReadUTC(&dateTime);
    TimePhase  timePhase = TIMEPHASE_UNKNOWN;
    if (DATETIME_IsValid(&dateTime))
    {   DSTMode dstMode;
        STORING_Load(STO_DstMode, &dstMode);
        switch (dstMode)
        {   case DSTMODE_OFF:
                timePhase = TIMEPHASE_STD;
                break;
            case DSTMODE_ON:
                timePhase = TIMEPHASE_DST;
                break;
            case DSTMODE_AUTO:
            {   switch (DATETIME_DSTPhase(&dateTime))
                {   case DST_MET:  timePhase = TIMEPHASE_STD; break;
                    case DST_UITA: timePhase = TIMEPHASE_STD; break;
                    case DST_MEST: timePhase = TIMEPHASE_DST; break;
                    case DST_UITS: timePhase = TIMEPHASE_DST; break;
                    default:       ;
                }
                break;
            }
            case DSTMODE_DCF:
                if      (BOARD_RadioTimeZone() == 1) timePhase = TIMEPHASE_STD;
                else if (BOARD_RadioTimeZone() == 2) timePhase = TIMEPHASE_DST;
                break;
            default:
                ;
        }
    }
    return timePhase;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichtenausstatter  ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupUTCBroadcast(Cy4Message* message)
{
    DateTime dateTime;
    BOARD_ReadUTC(&dateTime);
    if (DATETIME_IsValid(&dateTime))
    {   UInt08 data[7];
        DATETIME_ToData(&dateTime, data);
        CY4MESSAGE_SetupWrite(message, CY4ADDRESS_Create(0, 0), 0xE00010, data, 7);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupLTCBroadcast(Cy4Message* message)
{
    DateTime dateTime;
    ReadLocalTime(&dateTime);
    if (DATETIME_IsValid(&dateTime))
    {   UInt08 data[7];
        DATETIME_ToData(&dateTime, data);
        CY4MESSAGE_SetupWrite(message, CY4ADDRESS_Create(0, 0), 0xE00020, data, 7);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfragebearbeiter  ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: UniversalTime (read write)                                                  //
//                                                                                      //
//  Beschreibung: Zugriff auf die Weltzeit UTC.                                         //
//                                                                                      //
//               0   1   2   3   4   5   6                                              //
//             +---+---+---+---+---+---+---+                                            //
//  Belegung:  |YEh|YEl|MON|DAY|HUR|MIN|SEC|                                            //
//             +---+---+---+---+---+---+---+                                            //
//                                                                                      //
//      [YEx] Jahr:  1700 .. 9000 (FFFFh)                                               //
//      [MON] Monat:    1 ..   12 (FFh)                                                 //
//      [DAY] Tag:      1 ..   31 (FFh)                                                 //
//      [HUR] Stunde:   1 ..   24 (FFh)                                                 //
//      [MIN] Minute:   1 ..   59 (FFh)                                                 //
//      [SEC] Sekunde:  1 ..   59 (FFh)                                                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnUniversalTime(const Cy4Service* service, Cy4Message* message)
{
    DateTime dateTime;
    UInt08   data[7];
    BOARD_ReadUTC(&dateTime);
    if (!DATETIME_IsValid(&dateTime))
    {   DateTime silvester2000 = SILVESTER_2000;
        dateTime = silvester2000;
    }
    DATETIME_ToData(&dateTime, data);
    CY4SERVICE_TransferPayload(service, message, data);
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   DATETIME_FromData(data, &dateTime);
        BOARD_WriteUTC(&dateTime);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: LocalTime (read write)                                                      //
//                                                                                      //
//  Beschreibung: Zugriff auf die Ortszeit. Die Ortszeit wird aus der UTC unter         //
//                Berücksichtigung der Zeitzone und gegebenenfalls der Sommerzeit       //
//                (s. TimeZone) abgeleitet.                                             //
//                                                                                      //
//               0   1   2   3   4   5   6                                              //
//             +---+---+---+---+---+---+---+                                            //
//  Belegung:  |YEh|YEl|MON|DAY|HUR|MIN|SEC|                                            //
//             +---+---+---+---+---+---+---+                                            //
//                                                                                      //
//      [YEx] Jahr:  1700 .. 9000 (FFFFh)                                               //
//      [MON] Monat:    1 ..   12 (FFh)                                                 //
//      [DAY] Tag:      1 ..   31 (FFh)                                                 //
//      [HUR] Stunde:   1 ..   24 (FFh)                                                 //
//      [MIN] Minute:   1 ..   59 (FFh)                                                 //
//      [SEC] Sekunde:  1 ..   59 (FFh)                                                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnLocalTime(const Cy4Service* service, Cy4Message* message)
{
    DateTime dateTime;
    UInt08   data[7];
    ReadLocalTime(&dateTime);
    if (!DATETIME_IsValid(&dateTime))
    {   DateTime silvester2000 = SILVESTER_2000;
        dateTime = silvester2000;
    }
    DATETIME_ToData(&dateTime, data);
    CY4SERVICE_TransferPayload(service, message, data);
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   DATETIME_FromData(data, &dateTime);
        WriteLocalTime(&dateTime);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TimeZone (read write)                                                       //
//                                                                                      //
//  Beschreibung: Legt die Normalzeitzone der Ortszeit und die Sommerzeit-Betriebsart   //
//                fest. TZN bestimmt dabei die Zeitzone der Normalzeit als Differenz    //
//                zur UTC in Stunden, beispielsweise UTC+1 = MEZ).                      //
//                Grundlage der automatischen Zeitumstellung ist die mitteleuropäische  //
//                Sommerzeitregelung. Man beachte, dass in der Sommerzeitphase die      //
//                aktuelle Ortszeit um eine Stunde vorversetzt ist.                     //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |TZN|DST|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [TZN] Normalzeitzone der Ortszeit: -11 .. +12 [h], sonst Fehler. Voreingestellt     //
//        ist 1.                                                                        //
//  [DST] Sommerzeit-Betriebsart:                                                       //
//          0 = Aus: Normalzeit.                                                        //
//          1 = Ein: Sommerzeit, (Normalzeit+1 Stunde).                                 //
//          2 = Automatische Umschaltung gemäß MESZ-Vorgabe (Voreinstellung).           //
//          3 = Automatische Umschaltung gemäß empfangenen Funkzeit. Ist der            //
//              Funkempfang deaktiviert oder hat das System seit dem letzten Start noch //
//              Zeitinformation erhalten wird die Zeitphase wie bei 2 ermittelt.        //
//          x = Fehler.                                                                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTimeZone(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[2];
    SInt08 timeZone;
    STORING_Load(STO_TimeZone, &timeZone);
    data[0] = (UInt08)timeZone;
    DSTMode dstMode;
    STORING_Load(STO_DstMode, &dstMode);
    data[1] = (UInt08)dstMode;
    CY4SERVICE_TransferPayload(service, message, data);
    timeZone = (SInt08)data[0];
    dstMode  = (DSTMode)data[1];
    if (timeZone >= -11 && timeZone <= 12 && data[1] < 4)
    {   STORING_Save(&timeZone, STO_TimeZone);
        STORING_Save(&dstMode, STO_DstMode);
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TimePhase (read only)                                                       //
//                                                                                      //
//  Beschreibung: Informiert darüber, ob die aktuelle Ortzeit die Normalzeit oder die   //
//                Sommerzeit (gem. MESZ) ist. Ist die Uhrzeit nicht verfügbar, liefert  //
//                TPH den Wert 2.                                                       //
//                                                                                      //
//               0                                                                      //
//             +---+                                                                    //
//  Belegung:  |TPH|                                                                    //
//             +---+                                                                    //
//                                                                                      //
//      [TPH] Aktuelle Zeitphase: 0 = Normalzeit, 1 = Sommerzeit, 2 = Unbekannt.        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTimePhase(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    UInt08  data;
    switch (CurrentTimePhase())
    {   case TIMEPHASE_STD: data = 0; break;
        case TIMEPHASE_DST: data = 1; break;
        default:            data = 2;
    }
    CY4MESSAGE_Data(message)[0] = data;
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: RadioTimeEnable (read write)                                                //
//                                                                                      //
//  Beschreibung: Aktiviert oder deaktiviert den Empfang der Funkzeit.                  //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |RTE|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [RTE] Funkzeitempfang: 1 = Ein, 0 = Aus, x = Ohne Wirklung.                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnRadioTimeEnable(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data = (bool)BOARD_IsRadioTimeEnabled();
    CY4SERVICE_TransferPayload(service, message, &data);
    if (data < 2)
    {   bool enable = (bool)data;
        BOARD_SetRadioTimeEnabled(enable);
        STORING_Save(&enable, STO_RadioTimeEnabled);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: LastRadioTimeUpdate (read only)                                             //
//                                                                                      //
//  Beschreibung: Informiert darüber, wie viele Minuten seit dem letzten korrekten      //
//                Funkzeitempfang vergangen sind. Der Zähler läuft bis maximal 28800    //
//                (= 20 Tage) und bleibt danach auf diesem Wert stehen. Ist der Funk-   //
//                empfang deaktiviert oder hat das System seit dem letzten Start noch   //
//                keine Zeitinformation erhalten, ist [TUx] 8000h.                      //
//                                                                                      //
//               0   1                                                                  //
//             +---+---+                                                                //
//  Belegung:  |TU0|TU1|                                                                //
//             +---+---+                                                                //
//                                                                                      //
//  [TUx]: Minuten seit dem letzten Funkzeitempfang: 0 .. 28800 [min] oder 8000h.       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnLastRadioTimeUpdate(const Cy4Service* service, Cy4Message* message)
{
    SInt16 elapsed = BOARD_LastRadioTimeUpdate();
    UInt08  data[2] = { HiByte(elapsed), LoByte(elapsed) };
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ClocksGood (read only)                                                      //
//                                                                                      //
//  Beschreibung: Informiert über den Funktionszustand der Uhren. Der Empfangszustand   //
//                der Funkzeit gilt als "Gut", wenn mindestens 5 Sekunden lang verwert- //
//                bare Funksignale eingehen. Ist der Empfang der Funkzeit abgeschaltet, //
//                (vergl. RadioTimeEnable) ist [RCS] 0.                                 //
//                                                                                      //
//               0   1                                                                  //
//             +---+---+                                                                //
//  Belegung:  |RCS|RTC|                                                                //
//             +---+---+                                                                //
//                                                                                      //
//  [RCS]: 1 = Empfangszustand "gut", 0 = "schlecht" oder abgeschaltet.                 //
//  [RTC]: 1 = Echtzeituhr in Ordnung, 0 = Fehler.                                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnClocksGood(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[2];
    data[0] = BOARD_RadioTimeGood();
    data[1] = BOARD_RealTimeClockGood();
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BroadcastsEnable (read write)                                               //
//                                                                                      //
//  Beschreibung: Aktivierung der Serverzeit-Rundrufe. Die Rundrufe werden zu jeder     //
//                vollen Stunden an 000.000 getätigt. Übertragen wird das Datum und die //
//                die Zeit im gleichen Format wie bei UniversalTime oder LocalTime.     //
//                Dienste:                                                              //
//                              Rundruf Weltzeit: E00010h:07                            //
//                              Rundruf Ortszeit: E00020h:07                            //
//                                                                                      //
//               0   1                                                                  //
//             +---+---+                                                                //
//  Belegung:  |UTC|LCT|                                                                //
//             +---+---+                                                                //
//                                                                                      //
//  [UTC]: Weltzeit-Rundruf: 1 = Ausführen (Voreinstellung), 0 = Nicht ausführen.       //
//  [LCT]: Ortszeit-Rundruf: 1 = Ausführen (Voreinstellung), 0 = Nicht ausführen.       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBroadcastsEnable(const Cy4Service* service, Cy4Message* message)
{
    bool enabled[2];
    STORING_Load(STO_BroadcastEnabled, &enabled);
    UInt08 data[2];
    data[0] = enabled[0];
    data[1] = enabled[1];
    CY4SERVICE_TransferPayload(service, message, data);
    enabled[0] = (bool)data[0];
    enabled[1] = (bool)data[1];
    STORING_Save(&enabled, STO_BroadcastEnabled);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: SolarEvents (read only)                                                     //
//                                                                                      //
//  Beschreibung: Liefert die lokale Sonnenauf- und untergangszeiten sowie die Zeiten   //
//                der bürgerlichen Dämmerung für den aktuellen Tag (Ortszeit). Die      //
//                Berechnung basiert u.a. auf der Zeitzone und der Standortangabe in    //
//                Location.                                                             //
//                                                                                      //
//               0   1             4   5                                                //
//             +---+---+  2   3  +---+---+  6   7                                       //
//  Belegung:  |BMH|BMM|+---+---+|SSH|SSM|+---+---+                                     //
//             +---+---+|SRH|SRM|+---+---+|BAH|BAM|                                     //
//                      +---+---+         +---+---+                                     //
//                                                                                      //
//  [BMH]: Bürgerlichen Morgendämmerung, Stunde: 0 .. 23 [h], 255                       //
//  [BMM]: Bürgerlichen Morgendämmerung, Minute: 0 .. 59 [m], 255                       //
//  [SRH]: Sonnenaufgang, Stunde: 0 .. 23 [h], 255                                      //
//  [SRM]: Sonnenaufgang, Minute: 0 .. 59 [m], 255                                      //
//  [SSH]: Sonnenuntergang, Stunde: 0 .. 23 [h], 255                                    //
//  [SSM]: Sonnenuntergang, Minute: 0 .. 59 [m], 255                                    //
//  [BAH]: Bürgerlichen Abenddämmerung, Stunde: 0 .. 23 [h], 255                        //
//  [BAM]: Bürgerlichen Abenddämmerung, Minute: 0 .. 59 [m], 255                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnSolarEvents(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[8];

    data[0] = civilDawnTime.hour;
    data[1] = civilDawnTime.minute;
    data[2] = sunriseTime.hour;
    data[3] = sunriseTime.minute;
    data[4] = sunsetTime.hour;
    data[5] = sunsetTime.minute;
    data[6] = civilDuskTime.hour;
    data[7] = civilDuskTime.minute;
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: Location (read write)                                                       //
//                                                                                      //
//  Beschreibung: Zugriff auf die Standortdaten als Basis für die Sonnenereignisse in   //
//                SolarEvents. Voreingestellt ist nichts (8000h)                        //
//                                                                                      //
//               0   1                                                                  //
//             +---+---+  2   3                                                         //
//  Belegung:  |BGh|BGl|+---+---+                                                       //
//             +---+---+|LGh|LGl|                                                       //
//                      +---+---+                                                       //
//                                                                                      //
//  [BGx]: Geographische Breite des Standortes: -6500 .. +6500 [°÷100] oder 8000h.      //
//  [LGx]: Geographische Länge  des Standortes: -17999 .. +18000 [°÷100] oder 8000h.    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnLocation(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[4];

    SInt16  latitude;
    SInt16  longitude;
    STORING_Load(STO_Latitude, &latitude);
    STORING_Load(STO_Longitude, &longitude);
    data[0] = HiByte(latitude);
    data[1] = LoByte(latitude);
    data[2] = HiByte(longitude);
    data[3] = LoByte(longitude);
    CY4SERVICE_TransferPayload(service, message, data);
    latitude  = ToWord(data[0], data[1]);
    longitude = ToWord(data[2], data[3]);
    STORING_Save(&latitude, STO_Latitude);
    STORING_Save(&longitude, STO_Longitude);
    updateSolarEvent = true;
}


//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: LedSignaling (read write)                                                   //
//                                                                                      //
//  Beschreibung: Legt die Aufgabe der Info- und System-LEDs fest. Folgende Licht-      //
//                signale können eingestellt werden:                                    //
//                                                                                      //
//                0 = Kein Signal, Ausgeschaltet.                                       //
//                1 = Leuchtet durchgehend.                                             //
//                2 = Herzschlag-Signal.                                                //
//                3 = Leuchtet bei Busaktivität                                         //
//                4 = DCF-Empfang: LED ein/aus = korrekt/gestört (oder aus).            //
//                5 = Blinkt, wenn letzter Zeitempf. mehr als eine Woche zurückliegt.   //
//                6 = Blinkt bei Kommunikationsproblemen mit der Echtzeituhr.           //
//                x = Wirkungslos.                                                      //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |INF|SYS|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [INF] Lichtsignal der Info-LED. Werte siehe oben, Voreingestellt ist 0.             //
//  [SYS] Lichtsignal der System-LED. Werte siehe oben, Voreingestellt ist 0.           //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnLedSignaling(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[2];

    LedSignal signal[2];
    STORING_Load(STO_LedSignal, signal);
    data[0] = (UInt08)signal[0];
    data[1] = (UInt08)signal[1];
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] < 7) signal[0] = (LedSignal)data[0];
    if (data[1] < 7) signal[1] = (LedSignal)data[1];
    STORING_Save(signal, STO_LedSignal);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^').                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^').                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(&address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedBroadcasts(void)
{
    DateTime dateTime;
    BOARD_ReadUTC(&dateTime);
    if (!DATETIME_IsValid(&dateTime)) return;
    static UInt08 second = 0xFF;
    if (second != dateTime.time.second)
    {   second  = dateTime.time.second;
        if (dateTime.time.minute == 0)
        {   bool enabled[2];
            STORING_Load(STO_BroadcastEnabled, &enabled);
            if (second == 1 && enabled[0])
            {   CY4NET_Post(SetupUTCBroadcast, NULL);
            }
            if (second == 2 && enabled[1])
            {   CY4NET_Post(SetupLTCBroadcast, NULL);
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLedSignaling(void)
{
    LedSignal signal[2];
    STORING_Load(STO_LedSignal, signal);
    for (UInt08 i=0; i<2; i++)
    {   bool ledOn = false;
        switch (signal[i])
        {   case LEDSIGNAL_OFF:
                ledOn = false;
                break;
            case LEDSIGNAL_ON:
                ledOn = true;
                break;
            case LEDSIGNAL_HEARTBEAT:
            {   UInt16 ticks = BOARD_MSecTicks();
                ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180);
                break;
            }
            case LEDSIGNAL_BUS_ACTIVITY:
                ledOn = BOARD_BusActivity();
                break;
            case LEDSIGNAL_RADIO_GOOD:
                ledOn = BOARD_RadioTimeGood();
                break;
            case LEDSIGNAL_RADIO_BAD:
                ledOn = BOARD_IsRadioTimeEnabled() && !BOARD_RadioTimeGood();
                break;
            case LEDSIGNAL_DCF_FAIL:
            {   SInt16 elapsed = BOARD_LastRadioTimeUpdate();
                if (elapsed == INVALID || elapsed < 60*24*10)   // 10 Tage
                {   ledOn = false;
                }
                else
                {   ledOn = (bool)(BOARD_MSecTicks() & 0x100);
                }
                break;
            }
            case LEDSIGNAL_RTC_FAIL:
                if (BOARD_RealTimeClockGood())
                {   ledOn = false;
                }
                else
                {   ledOn = (bool)(BOARD_MSecTicks() & 0x100);
                }
                break;
            default:
                ledOn = false;
        }
        BOARD_SetLed((LED)i, ledOn);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Zustandsautomat Berechnung Sonnenereignisse                                         //
//                                                                                      //
//            +---------------------------------+                                       //
//       +--->|              IDLE               |                                       //
//       |    +-----------------------------+---+                                       //
//       |       /|\                        | UPDATE                           			//
//       |        |                         |  -> StartTimer()                          //
//       |        |                         |                                           //
//       |        |                         |                                           //
//       |        |                         |     +-------------------------+           //
//       |        |                         |     |                         |           //
//       |        |  RESET                 \|/   \|/      UPDATE            |           //
//       |        |   -> ClearTimes() +-----------------+  -> StartTimer()  |           //
//       |        +-------------------+ CALC_CIVIL_DAWN +-------------------+           //
//       |        |                   +--------+--------+                   |           //
//       |        |                            | TIMEOUT                    |           //
//       |        |                            |  -> StartTimer()           |           //
//       |        |                            |                            |           //
//       |        |  RESET                    \|/         UPDATE            |           //
//       |        |   -> ClearTimes() +-----------------+  -> StartTimer()  |           //
//       |        +-------------------+  CALC_SUNRISE   +-------------------+           //
//       |        |                   +--------+--------+                   |           //
//       |        |                            | TIMEOUT                    |           //
//       |        |                            |  -> StartTimer()           |           //
//       |        |                            |                            |           //
//       |        |  RESET                    \|/         UPDATE            |           //
//       |        |   -> ClearTimes() +-----------------+  -> StartTimer()  |           //
//       |        +-------------------+   CALC_SUNSET   +-------------------+           //
//       |        |                   +--------+--------+                   |           //
//       |        |                            | TIMEOUT                    |           //
//       |        |                            |  -> StartTimer()           |           //
//       |        |                            |                            |           //
//       |        |  RESET                    \|/         UPDATE            |           //
//       |        |   -> ClearTimes() +-----------------+  -> StartTimer()  |           //
//       |        +-------------------+ CALC_CIVIL_DUSK +-------------------+           //
//       |                            +--------+--------+                               //
//       |                                     | TIMEOUT                                //
//       |                                     |                                        //
//       |                                     |                                        //
//       +-------------------------------------+                                        //
//                                                                                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSolarEvents(void)
{
	typedef enum
	{	IDLE,
		CALC_CIVIL_DAWN,
		CALC_SUNRISE,
		CALC_SUNSET,
		CALC_CIVIL_DUSK
	} State;

	typedef enum
	{	UPDATE,
		RESET,
		NONE
	} Event;
	    	
	static State  state        = IDLE;	
	static Timer  timer        = TIMER_MILLISECONDS(1234);

	// Ereignisgenerator
	static SInt16 refLatitude  = INVALID;
	static SInt16 refLongitude = INVALID;
	static Date   refDate      = DATE_INVALID;
	Event event = NONE;
    SInt16  latitude;
    STORING_Load(STO_Latitude, &latitude);	
    SInt16  longitude;
    STORING_Load(STO_Longitude, &longitude);
	DateTime dateTime;
	ReadLocalTime(&dateTime);
	if (latitude == INVALID || longitude == INVALID || !DATETIME_IsValid(&dateTime))
	{	// Verwerfen, wenn das lokale Datum oder der Standort ungültig ist
		event = RESET;
	}
	else if (refLatitude  != latitude  ||
	         refLongitude != longitude ||
		     dateTime.date.month != refDate.month ||
			 dateTime.date.day   != refDate.day)
	{	refLatitude  = latitude;
		refLongitude = longitude;
		refDate      = dateTime.date;
		// Aktualisieren, wenn sich das lokale Datum oder der Standort verändert hat
		event = UPDATE;
	}		
	// Zustandsübergänge
	if (state == IDLE)
	{	if (event == UPDATE)
		{	TIMER_Restart(&timer);
			state = CALC_CIVIL_DAWN;
		}
	}
	else // Alle anderen Zustände
	{	if (event == RESET)
		{	const DateTime invalidDateTime = DATETIME_INVALID;
			refDate       = invalidDateTime.date;
			refLatitude   = INVALID;
			refLongitude  = INVALID;
			civilDawnTime = invalidDateTime.time;
			sunriseTime   = invalidDateTime.time;
			sunsetTime    = invalidDateTime.time;
			civilDuskTime = invalidDateTime.time;			
			state = IDLE;
		}
		else if (event == UPDATE)
		{	state = CALC_CIVIL_DAWN;
			TIMER_Restart(&timer);
		}
		else if (TIMER_HasExpired(&timer))
		{	switch (state)
			{	case CALC_CIVIL_DAWN:
					civilDawnTime = CalcSolarTime(SOLAREVENT_CIVIL_DAWN, latitude, longitude, &dateTime.date);
					state = CALC_SUNRISE;
					break;
				case CALC_SUNRISE:
					sunriseTime = CalcSolarTime(SOLAREVENT_SUNRISE, latitude, longitude, &dateTime.date);
					state = CALC_SUNSET;
					break;
				case CALC_SUNSET:
					sunsetTime = CalcSolarTime(SOLAREVENT_SUNSET, latitude, longitude, &dateTime.date);
					state = CALC_CIVIL_DUSK;
					break;
				case CALC_CIVIL_DUSK:
					civilDuskTime = CalcSolarTime(SOLAREVENT_CIVIL_DUSK, latitude, longitude, &dateTime.date);
					state = IDLE;															
					break;
				default:
					state = IDLE;
			}
			TIMER_Restart(&timer);
		}				
	}    
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                    service                     |        handler        |
        //+-----------------------------------------.....--+-----------------------+
        { CY4SERVICE(0x700010,  7  , CY4SERVICE_READ_WRITE), OnUniversalTime       },
        { CY4SERVICE(0x700020,  7  , CY4SERVICE_READ_WRITE), OnLocalTime           },
        { CY4SERVICE(0x700030,  2  , CY4SERVICE_READ_WRITE), OnTimeZone            },
        { CY4SERVICE(0x700040,  1  , CY4SERVICE_READ_ONLY) , OnTimePhase           },
        { CY4SERVICE(0x700050,  1  , CY4SERVICE_READ_WRITE), OnRadioTimeEnable     },
        { CY4SERVICE(0x700060,  2  , CY4SERVICE_READ_ONLY) , OnLastRadioTimeUpdate },
        { CY4SERVICE(0x700070,  2  , CY4SERVICE_READ_ONLY) , OnClocksGood          },
        { CY4SERVICE(0x700080,  2  , CY4SERVICE_READ_WRITE), OnBroadcastsEnable    },
        { CY4SERVICE(0x700100,  8  , CY4SERVICE_READ_ONLY) , OnSolarEvents         },
        { CY4SERVICE(0x700120,  4  , CY4SERVICE_READ_WRITE), OnLocation            },
        { CY4SERVICE(0x700400,  2  , CY4SERVICE_READ_WRITE), OnLedSignaling        },
        { CY4SERVICE(0xFFFF00, 61  , CY4SERVICE_READ_WRITE), OnDeviceLabel         },
        { CY4SERVICE(0xFFFF40,  1  , CY4SERVICE_WRITE_ONLY), OnFactorySettings     },
        { CY4SERVICE(0xFFFFF0,  7  , CY4SERVICE_READ_ONLY ), OnCallTypePlate       },
        { CY4SERVICE(0xFFFFF8,  1  , CY4SERVICE_WRITE_ONLY), OnEnterApplication    },
        { CY4SERVICE(0xFFFFFA,  1  , CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader  },
        { CY4SERVICE(0xFFFFFC,  2  , CY4SERVICE_WRITE_ONLY), OnChangeCy4Address    }
    };
    CY4NET_AssignServices(servicesTable);
    Cy4Address address = STORING_LoadCy4SelfAdr();
    CY4NET_SetAddress(address);
    bool radioTimeEnabled;
    STORING_Load(STO_RadioTimeEnabled, &radioTimeEnabled);
    BOARD_SetRadioTimeEnabled(radioTimeEnabled);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProceedBroadcasts();
    ProceedLedSignaling();
    ProceedSolarEvents();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
