﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/twi.h>
#include "AVRChip.h"
#include "Board.h"
#include "Timer.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define SYS_LED         B,2
#define DCF_PIN         D,5
#define INFO_LED        D,6
#define I2C_ADR_DS1307  0x68

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,             // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,        // · Starte den Programmlader
    DO_PROCEED,             // · Führe die Arbeit fort
    DO_SHUTDOWN             // · Schalte das Board ab
} DoNext;

typedef struct              // Bus-Übertragungsstatus:
{   bool   pending : 1;     // · Übertragung im Gange
    bool   echo    : 1;     // · Echo erhalten
    bool   error   : 1;     // · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap : 5;     // · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct              // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;     // · Zeitüberschreitung bei <byteGap> aufgetreten
    UInt08 padding : 7;     // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;     // · Datenbyte
} BusRxItem;

typedef struct              // Zustandsinformationen des Funkzeitempfangs:
{   UInt08    bitCount;     // · Empfangsbitzähler (aktuell empfangene Sekunde): 0 .. 59.
    UInt08    goodBits;     // · Sequenzlänge erfolgreich empf. Bits: 0 .. 250.
    SInt16    elapsed;      // · Zeit, seit dem letzten Empf.: 0 .. 28800 [min], INVALID.
    SInt08    timeZone;     // · Empfangene Zeitzone: (UTC+)1, (UTC+)2 oder -128.
} DCFStatus;

typedef enum                // Ereignisse der DCF-PWM Abtastung
{   DCF_0    = '0',         // · Null erkannt
    DCF_1    = '1',         // · Eins erkannt
    DCF_MARK = 'm',         // · Minutenanfang erkannt
    DCF_FAIL = 'e',         // · Abtastfehler
    DCF_NONE = 'x'          // · Kein Ereignis
} DCFEvent;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16      msecTicks                             = 0;
static volatile UInt16      secTicks                              = 0;
static volatile UInt08      busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08      busTxQueueHead                        = 0;
static volatile UInt08      busTxQueueLength                      = 0;
static volatile UInt08      busLoopbackByte                       = 0;
static volatile BusStatus   busStatus                             = { false, false, false, 0};
static volatile BusRxItem   busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { false } };
static volatile UInt08      busRxQueueHead                        = 0;
static volatile UInt08      busRxQueueTail                        = 0;
static          UInt08      busActivityIndicator                  = 0;
static          bool        radioTimeEnabled                      = false;
static volatile DCFEvent    dcfEvent                              = DCF_NONE;
static          DCFStatus   dcfStatus                             = { 0, 0, INVALID, -128 };
static          DateTime    sysDateTime                           = DATETIME_INVALID;
static          bool        rtClockFailure                        = false;
static          DoNext      doNext                                = DO_STARTUP;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();            // Interrupts sperren
    wdt_enable(WDTO_15MS);          // Watchdog-Intervall auf kürzeste Zeit setzen
    __asm__ volatile
    (   "ldi r30, lo8 (%0) \n"      // Nach einem Neustart des Systems wird als
        "ldi r31, hi8 (%0) \n"      // erstes das Programmladermodul ausgeführt.
        "ldi r24, %1       \n"      // Der Programmlader nutzt zur Steuerung seines
        "ldi r25, hi8 (%2) \n"      // Ablaufs Variablen, die in der "Noinit"-Sektion
        "0:                \n"      // angesiedelt sind, d.h. sie werden im Rahmen
        "st  Z+,  r24      \n"      // eines Neustartes nicht initialisiert.
        "cpi r30, lo8 (%2) \n"      // Um nun zu vermeiden, dass übrig gebliebene
        "cpc r31, r25      \n"      // Werte des SRAMs unerwünschte Effekte im
        "brlo 0b"                   // Programmlader hervorrufen, wird vor dem
            :                       // Reset das komplette SRAM mit Nullen überschrieben.
            : "i" (RAMSTART),       // Da durch die Nullung Heap wie Stack komplett
              "i" (0),              // zerstört sind, ist diese Aktion ein Punkt-Ohne-
              "i" (RAMEND+1)        // Wiederkehr.
    );
    while (true);                   // Warten, bis Watchdog-Reset eintritt
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterPgmLdr(void)
{
    if (pgm_read_word(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1) == 0x940C)
    {   GOTO(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1 + 4);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
    MCUSR = 0;
    wdt_disable();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TransceiveI2C(): Sendet als erstes <writeCount> Datenbytes aus <writeBuffer> an das //
//  ---------------  I²C-Gerät an Adresse <address> und liest danach <readCount> Daten- //
//  bytes wieder zurück. Die empfangenen Datenbytes werden in <readBuffer> abgelegt.    //
//  Die Schreib-Lese-Sequenz wird auf auf dem I²C-Bus durch einen "RestartRead"-Zustand //
//  getrennt.                                                                           //
//  Da Schreib- und Lesevorgänge nacheinander durchgeführt werden, können <writeBuffer> //
//  und <readBuffer> den gleichen Speicherbereich referenzieren.                        //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127.                              //
//           · writeBuffer: Speicherbereich mit zu schreibenden Bytes.                  //
//           · writeCount:  Anzahl Bytes in <writeBuffer>.                              //
//           · readBuffer:  Speicherbereich, der die zu lesenden Bytes aufnimmt.        //
//           · readCount:   Anzahl zu lesender Bytes.                                   //
//  Ausgabe: 'true': Schreib-Lese-Vorgang war erfolgreich.                              //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
TransceiveI2C(SInt08          address,
                const UInt08* writeBuffer,
                UInt08        writeCount,
                UInt08*       readBuffer,
                UInt08        readCount)
{
    bool success = true;
    // Schreiben
    if (writeCount > 0)
    {   success = false;
        // I²C Einheit einschalten und Start-Kennung senden
        TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);
        UInt08 t;
        for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us(); // max. 1ms warten
        if (t > 0)
        {   if (TW_STATUS == TW_START)
            {   // Zielgeräte-Addresse mit Schreib-Anforderung senden (SLA+W)
                TWDR = ((UInt08)address << 1) | TW_WRITE;
                TWCR = BV(TWINT)|BV(TWEN);
                for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
                if (t > 0)
                {   if (TW_STATUS == TW_MT_SLA_ACK)
                    {   // <writeCount> Datenbytes übertragen
                        UInt08 i;
                        for (i=0; i<writeCount; i++)
                        {   TWDR = writeBuffer[i];
                            TWCR = BV(TWINT)|BV(TWEN);
                            for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
                            if (t == 0 || TW_STATUS != TW_MT_DATA_ACK) break;
                        }
                        success = i == writeCount;
                    }
                }
            }
        }
    }
    // Lesen
    if (readCount > 0 && success)
    {   success = false;
		// Start/Repeated start-Kennung senden
        TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);
        UInt08 t;
        for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us(); // max. 1ms warten
        if (t > 0)
        {   UInt08 status;
            if (writeCount == 0) status = TW_START;
            else                 status = TW_REP_START;
            if (TW_STATUS == status)
            {   // Zielgeräte-Addresse mit Leseanforderung übertragen (SLA+R)
                TWDR = ((UInt08)address << 1) | TW_READ;
                TWCR = BV(TWINT) | BV(TWEN);
                for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
                if (t > 0)
                {    if (TW_STATUS == TW_MR_SLA_ACK)
                    {   // <readCount> Datenbytes entgegennehmen, das letzte unbestätigt (NACK)
                        UInt08 i;
                        for (i=0; i<readCount; i++)
                        {   UInt08  acknowledge;
                            if (i < readCount-1)
                            {   TWCR = BV(TWINT) | BV(TWEN) | BV(TWEA);
                                acknowledge = TW_MR_DATA_ACK;
                            }
                            else
                            {   TWCR = BV(TWINT) | BV(TWEN);
                                acknowledge = TW_MR_DATA_NACK;
                            }
                            for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
                            if (t == 0 || TW_STATUS != acknowledge) break;
                            readBuffer[i] = TWDR;
                        }
                        success = i == readCount;
                    }
                }
            }
        }
    }
    // Stop-Kennung senden und I²C Einheit abschalten
    TWCR = BV(TWINT)|BV(TWSTO)|BV(TWEN);
    Freeze100us();
    TWCR = 0;
    return success;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TransmitI2C(): Sendet <count> Datenbytes aus <buffer> an das I²C-Gerät an Adresse   //
//  -------------  <address>.                                                           //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127.                              //
//           · buffer:  Speicherbereich mit zu schreibenden Bytes.                      //
//           · count:   Anzahl Bytes in <buffer>.                                       //
//  Ausgabe: 'true': Schreibvorgang war erfolgreich.                                    //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
TransmitI2C(SInt08 address, const UInt08* buffer, UInt08 count)
{
    return TransceiveI2C(address, buffer, count, NULL, 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ReceiveI2C(): Liest <count> Datenbytes von dem I²C-Gerät an Adresse <address> und   //
//  ------------  legt die empfangenen Bytes in <buffer> ab.                            //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127                               //
//           · buffer:  Speicherbereich, der die zu lesenden Bytes aufnimmt.            //
//           · count:   Anzahl zu lesender Bytes.                                       //
//  Ausgabe: 'true': Lesevorgang war erfolgreich.                                       //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
ReceiveI2C(SInt08 address, UInt08* buffer, UInt08 count)
{
    return TransceiveI2C(address, NULL, 0, buffer, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DS1307_Read(): Liest die Echtzeituhr vom Typ DS1307 aus. Schlägt das Auslesen fehl, //
//  -------------  kehrt die Funktion mit False zurück. Datum und Zeit in <dateTime>    //
//  sind in dem Fall ungültig. Kehrt die Funktion mit 'true' zurück, aber Datum und     //
//  Zeit in <dateTime> sind dennoch ungültig, steht die Echtzeituhr.                    //
//                                                                                      //
//  Eingabe: · dateTime: Zeiger auf eine DateTime-Struktur, die das Datum und die Zeit  //
//                       aufnimmt (darf nicht NULL sein).                               //
//  Ausgabe: 'true', wenn die Echtzeituhr ausgelesen werden konnte.                     //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
DS1307_Read(DateTime* dateTime)
{
    UInt08    data[7];
    DateTime  invalidDateTime = DATETIME_INVALID;

    *dateTime = invalidDateTime;
    data[0] = 0;
    if (!TransceiveI2C(I2C_ADR_DS1307, data, 1, data, CARDINALITY(data))) return false;
    if (data[0]&0x80) return true;                                      // "clock halt"?
    dateTime->time.second = ((data[0]>>4)&0x0F)*10 + (data[0]&0x0F);    // Sekunde
    dateTime->time.minute = ((data[1]>>4)&0x0F)*10 + (data[1]&0x0F);    // Minute
    dateTime->time.hour   = ((data[2]>>4)&0x0F)*10 + (data[2]&0x0F);    // Stunde
    dateTime->date.day    = ((data[4]>>4)&0x0F)*10 + (data[4]&0x0F);    // Tag
    dateTime->date.month  = ((data[5]>>4)&0x0F)*10 + (data[5]&0x0F);    // Monat
    dateTime->date.year   = 2000 + ((data[6]>>4)&0x0F)*10 + (data[6]&0x0F);  // 2000+Jahr
    return DATETIME_IsValid(dateTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DS1307_Write(): Stellt die Hardwareuhr vom Typ DS1307 und setzt sie gegebenenfalls  //
//  --------------  in Gang. Sind die in <dateTime> übergebenen Informationen ungültig, //
//  wird die Hardwareuhr angehalten.                                                    //
//                                                                                      //
//  Eingabe: · dateTime: Zeiger auf eine DateTime-Struktur, die das zu stellende Datum  //
//                       und die Zeit bereitstellt (darf nicht NULL sein).              //
//  Ausgabe: 'true', wenn der Zugiff auf die Echtzeituhr erfolgreich war.               //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
DS1307_Write(const DateTime* dateTime)
{
    UInt08 data[9];

    data[0] = 0;
    if (DATETIME_IsValid(dateTime))
    {   data[1] = (dateTime->time.second/10<<4)+(dateTime->time.second%10);
        data[2] = (dateTime->time.minute/10<<4)+(dateTime->time.minute%10);
        data[3] = (dateTime->time.hour/10<<4)+(dateTime->time.hour%10); // 24h-Darstellung
        data[4] = DATE_WeekDay(&dateTime->date);                        // Wochentag
        data[5] = (dateTime->date.day/10<<4)+(dateTime->date.day%10);
        data[6] = ((dateTime->date.month)/10<<4)+((dateTime->date.month)%10);
        data[7] = ((dateTime->date.year%100)/10<<4)+(dateTime->date.year%10);  // Kein Jahrhundert
        data[8] = 0x10;                                             // Square-Wave 1Hz ein
        return TransmitI2C(I2C_ADR_DS1307, data, CARDINALITY(data));
    }
    else
    {   data[1] = 0x80;
        return TransmitI2C(I2C_ADR_DS1307, data, 2);  // Uhr (nur) anhalten
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER2_COMPA_vect)
{
    // Abweichungskompensation              // Quarztaktbedingt findet der IRQ-Aufruf
    {   static UInt08 counter = 0;          // bei einem Zeitgeber-Zählerwert von 115
        if (counter == 4)                   // nicht exakt jede Millisekunde statt,
        {   counter = 0;                    // sondern etwas schneller:
            OCR2A = F_CPU/(32*1000UL);      //
        }                                   //      115 ÷ 115.2 = 0.9983 ms
        else                                //
        {   counter++;                      // Zwecks Ausgleichs dieser Ungenauigkeit
            OCR2A = F_CPU/(32*1000UL)-1;    // wird bei jedem fünften Aufruf, der
        }                                   // Überlauf um einen Zählerwert verlängert:
    }                                       // 4·(115÷115.2) + (116÷115.2) = 5 ms
    // Ticker inkrementieren
    {   msecTicks++;
        static UInt16 counter = 0;
        counter++;
        if (counter == 1000)
        {   counter = 0;
            secTicks++;
        }
    }
    // Bus Byte-Zeitlücke
    if (busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT)
    {   busStatus.byteGap++;
        if (busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT)
        {   busRxQueue[busRxQueueTail].timeout = true;
        }
    }
    // Abtastung des DCF77 PWM-Signals
    {   static UInt08 counter = 0;
        counter++;
        if (counter == 10)
        {   // 10 ms - Zeitintervall
            counter =  0;
            //////////////////////////////////////////////////////////////////////////////////////
            //                                                                                  //
            //  Mealy-Automat zur DCF77 PWM Abtastung:                                          //
            //  --------------------------------------                                          //
            //                                                                                  //
            //                                                                                  //
            //   Darstellung '0':    Abtastung '0':                                             //
            //                                                                                  //
            //     |                 Hi ··+-----------+    :      :      :      :      :        //
            //   1·+---+                  |//////:////|    :      :      :      :      :        //
            //     |///|              S0  |/ S1 /: S2 | S3 :  S4  :  S5  :  S6  :  S11 :        //
            //     |///|                  |//////:////|    :      :      :      :      :        //
            //   0·+---+---+--->     Lo ··+//////:////+---------------------------------···     //
            //        100     ms     -----+-----------------------+------------------------>    //
            //                            :      :         :      :      :      :      :   t    //
            //                            0      T0        T1     T2     T3     T4     T5       //
            //                                                                                  //
            //   Darstellung '1':    Abtastung '1':                                             //
            //                                                                                  //
            //     |                 Hi ··+-------------------------+    :      :      :        //
            //   1·+-------+              |//////://////://////:////|    :      :      :        //
            //     |///////|          S0  |/ S1 /:/ S2 /:/ S7 /: S8 | S9 :  S10 :  S11 :        //
            //     |///////|              |//////://////://////:////|    :      :      :        //
            //   0·+---+---+--->     Lo ··+//////://////://////:////+-------------------···     //
            //            200 ms     -----+--------------------+--------------------------->    //
            //                            :      :      :      :         :      :      :    t   //
            //                            0      T0     T1     T2        T3     T4     T5       //
            //                                                                                  //
            //                                                                                  //
            //   Mealy-Automat M:                                                               //
            //                                                                                  //
            //   M = { S,I,O,zt,s0,F }                             +------+                     //
            //                                                +--->|//////|                     //
            //     S  = { S0, .. ,S12 }                  Lo,x |    |/ S0 /|                     //
            //     I  = { Ti,Hi,Lo  }                         +----+//////|                     //
            //     O  = { x,a,0,1,m,e }                            +----+-+        Lo,x         //
            //     zt : S×I -> S×O                                      |          +---+        //
            //     s0 = S0                                     +-----+  | Hi,a     |   |        //
            //     F  = Ø                                 Hi,x |    \|/\|/         |  \|/       //
            //                                                 |   +------+      +-+-----+      //
            //                                                 +---+//////| Hi,m |///////|      //
            //   Eingaben I:                                       |/ S1 /|<-----+/ S12 /|<--+  //
            //     Ti = Zeitpunkt i erreicht    +----------------->|//////|      |///////|   |  //
            //     Hi = High-Signal liegt vor   |                  +--+---+      +-------+   |  //
            //     Lo = Low-Signal              |     Lo,x            |             Hi,x     |  //
            //                                  |     +--+            |             +--+     |  //
            //     Beispielzeiten für Ti:       |     |  |       Ti,x |             |  |     |  //
            //       T0 =   60 ms               |     | \|/          \|/           \|/ |     |  //
            //       T1 =  150 ms               |   +-+----+       +------+       +----+-+   |  //
            //       T2 =  170 ms               |   |//////|  Lo,x |//////| Ti,x  |//////|   |  //
            //       T3 =  250 ms               |   |/ S3 /|<------+/ S2 /+------>|/ S7 /|   |  //
            //       T4 =  900 ms               |   |//////|       |//////|       |//////|   |  //
            //       T5 = 1200 ms               |   +--+---+       +----+-+       +---+--+   |  //
            //       T6 = 2200 ms               |      |            /|\ |             |      |  //
            //                                  |      | Ti,x        |  |        Ti,x |      |  //
            //   Ausgaben O:                    |     \|/            +--+            \|/     |  //
            //     x = Kein Ereignis            |   +------+         Hi,x         +------+   |  //
            //     a = Beginn                   |   |//////|<---+            +--->|//////|   |  //
            //     0 = '0' erkannt              |   |/ S4 /|    | Lo,x  Hi,x |    |/ S8 /|   |  //
            //     1 = '1' erkannt              |   |//////+----+            +----+//////|   |  //
            //     m = Minutenanfang erkannt    |   +--+---+                      +---+--+   |  //
            //     e = Abtastfehler             |      |                              |      |  //
            //                                  |      | Ti,x                    Lo,x |      |  //
            //                                  |     \|/                            \|/     |  //
            //   Zustandsübergangsfunktion zt:  |   +------+                      +------+   |  //
            //                                  |   |//////|<---+            +--->|//////|   |  //
            //        |  Ti  |  Hi  |  Lo  |    |   |/ S5 /|    | Lo,x  Lo,x |    |/ S9 /|   |  //
            //    ----+------+------+------+    |   |//////+----+            +----+//////|   |  //
            //     S0 | S0,e | S1,a | S0,x |    |   +--+---+        Lo,x          +---+--+   |  //
            //     S1 | S2,x | S1,x | S0,e |    |      |            +---+             |      |  //
            //     S2 | S7,x | S2,x | S3,x |    |      | Ti,x       |   |        Ti,x |      |  //
            //     S3 | S4,x | S0,e | S3,x |    |     \|/           |  \|/           \|/     |  //
            //     S4 | S5,x | S0,e | S4,x |    |   +------+      +-+-----+       +-------+  |  //
            //     S5 | S6,x | S0,e | S5,x |    |   |//////| Ti,0 |///////|  Ti,1 |///////|  |  //
            //     S6 |S11,0 | S0,e | S6,x |    |   |/ S6 /+----->|/ S11 /|<------+/ S10 /|  |  //
            //     S7 | S8,x | S7,x | S0,e |    |   |//////|      |///////|       |///////|  |  //
            //     S8 | S0,e | S8,x | S9,x |    |   +-+----+      +-+---+-+       +-+-----+  |  //
            //     S9 |S10,x | S0,e | S9,x |    |     | /|\         |   |           | /|\    |  //
            //    S10 |S11,1 | S0,e |S10,x |    |     |  | Lo,x     |   |      Lo,x |  |     |  //
            //    S11 |S12,x | S1,a |S11,x |    |     +--+          |   |           +--+     |  //
            //    S12 | S0,e | S1,m |S12,x |    |              Hi,a |   | Ti,x               |  //
            //    ----+------+------+------+    +-------------------+   +--------------------+  //
            //                                                                                  //
            //////////////////////////////////////////////////////////////////////////////////////
            typedef struct
            {   UInt08  state;
                char    output;
            } Entry;

            static const Entry zeta[][3] PROGMEM =
            {   { { 0,'e'},{ 1,'a'},{ 0,'x'} },         //  Eingaben:
                { { 2,'x'},{ 1,'x'},{ 0,'e'} },         //   0 = Zeitpunkt i erreicht
                { { 7,'x'},{ 2,'x'},{ 3,'x'} },         //   1 = DCF-High-Signal liegt vor
                { { 4,'x'},{ 0,'e'},{ 3,'x'} },         //   2 = DCF-Low-Signal  liegt vor
                { { 5,'x'},{ 0,'e'},{ 4,'x'} },
                { { 6,'x'},{ 0,'e'},{ 5,'x'} },         //  Ausgaben:
                { {11,'0'},{ 0,'e'},{ 6,'x'} },         //   x = Kein Ereignis
                { { 8,'x'},{ 7,'x'},{ 0,'e'} },         //   a = Beginn
                { { 0,'e'},{ 8,'x'},{ 9,'x'} },         //   0 = '0' erkannt
                { {10,'x'},{ 0,'e'},{ 9,'x'} },         //   1 = '1' erkannt
                { {11,'1'},{ 0,'e'},{10,'x'} },         //   m = Minutenanfang erkannt
                { {12,'x'},{ 1,'a'},{11,'x'} },         //   e = Abtastfehler
                { { 0,'e'},{ 1,'m'},{12,'x'} }
            };

            static UInt08  state = 0;
            static UInt08  time  = 0;

            UInt08 in;
            if (time == CFG_DCF_T0 ||
                time == CFG_DCF_T1 ||
                time == CFG_DCF_T2 ||
                time == CFG_DCF_T3 ||
                time == CFG_DCF_T4 ||
                time == CFG_DCF_T5 ||
                time == CFG_DCF_T6)   in = 0;
            else if (GetPin(DCF_PIN)) in = 1;
            else                      in = 2;
            char out = pgm_read_byte(&(zeta[state][in].output));
            state = pgm_read_byte(&(zeta[state][in].state));
            if (out == 'm') time = 0;
            if      (out == 'a') time     = 0;
            else if (out != 'x') dcfEvent = (DCFEvent)out;
            time++;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART_TX_vect)
{                                   // IRQ Abfolge sowie die Signale <echo> und <pending>:
    if (!busStatus.echo)            //
    {   busStatus.error = true;     // Start Tx:    |1|
        busTxQueueLength = 0;       //               :
    }                               // Tx-IRQ:  ···--:-----|2|----|3|-···-|n|----| |-···
    busStatus.echo = false;         // Rx-IRQ:  ···--:-|1|--:-|2|--:--···-- -|n|--:--···
    UInt08 head = busTxQueueHead;   //               :  :___:  :___:          :___:
    if (busTxQueueLength == 0)      // echo:    ...__:__|   |__|   |____...___|   |__...
    {   busStatus.pending = false;  //               :____________________________:
    }                               // pending: ...__|                            |__...
    else
    {   busLoopbackByte = busTxQueue[head];
        UDR0            = busTxQueue[head];
        head++;
        if (head == CARDINALITY(busTxQueue)) head = 0;
        busTxQueueLength--;
    }
    busTxQueueHead = head;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART_RX_vect)
{
    UInt08 byte = UDR0;
    busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
    if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo
    // Byte in Empfangsschlange stellen
    UInt08 tail = busRxQueueTail;
    busRxQueue[tail].data = byte;
    tail++;
    if (tail == CARDINALITY(busRxQueue)) tail = 0;
    if (tail == busRxQueueHead) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
    else                        busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedDcfDecoding(void)
{
    static const UInt08  BCD[]    = { 1, 2, 4, 8, 10, 20, 40, 80 };
    static bool          parity   = false;
    static UInt08        minute   = 0;       // 0 .. 59
    static UInt08        hour     = 0;       // 0 .. 23
    static UInt08        day      = 1;       // 1 .. 31
    static UInt08        month    = 1;       // 1 .. 12
    static UInt08        year     = 0;       // (2000+) 0 .. 255
    static bool          dlsTime  = false;   // 'true' = empfangene Zeit war Sommerzeit

    if (!radioTimeEnabled)
    {   dcfStatus.bitCount = 0;
        dcfStatus.goodBits = 0;
        dcfStatus.elapsed  = INVALID;
        dcfStatus.timeZone = -128;
        return;
    }
    // Zeitzähler der DCF-Funkausfalldauer inkrementieren
    static Timer elapsedTimer = TIMER_SECONDS(60);
    if (TIMER_RestartIfExpired(&elapsedTimer))
    {   if (dcfStatus.elapsed != INVALID && dcfStatus.elapsed < 28800)
        {   dcfStatus.elapsed++;
        }
    }
    // Ereignis aus der PWM-Auswertung holen
    DCFEvent  event;
    Atomic
    {   event    = dcfEvent;
        dcfEvent = DCF_NONE;
    }
    SInt08 bitNo = (SInt08)dcfStatus.bitCount;      // 0 .. 59
    if (event == DCF_MARK && bitNo == 59)
    {   // Neue Zeit komplett
        DateTime  dateTime;
        dateTime.time.second = 0;
        dateTime.time.minute = minute;
        dateTime.time.hour   = hour;
        dateTime.date.day    = day;
        dateTime.date.month  = month;
        dateTime.date.year   = 2000 + year;
        if (DATETIME_IsValid(&dateTime))
        {   SInt08  shiftHrs;
            if (dlsTime) shiftHrs = -2;     // MESZ -> UTC
            else         shiftHrs = -1;     // MEZ  -> UTC
            DATETIME_ShiftHour(&dateTime, shiftHrs);
            sysDateTime       = dateTime;
            dcfStatus.elapsed = 0;
            TIMER_Restart(&elapsedTimer);
        }
        dcfStatus.bitCount = 0;     // Bereit für nächsten Zeitempfang
        dcfStatus.timeZone = 1+(SInt08)dlsTime;
        return;
    }
    bool dcfBit = false;
    switch (event)
    {   case DCF_NONE: return;
        case DCF_0:    dcfBit = false; break;
        case DCF_1:    dcfBit = true;  break;
        default:       bitNo  = -1;
    }
    parity ^= dcfBit;
    switch (bitNo)
    {   case  0:                                                // -+
        case  1:                                                //  |
        case  2:                                                //  |
        case  3:                                                //  |
        case  4:                                                //  |
        case  5:                                                //  |
        case  6:                                                //  |
        case  7:                                                //  | Bits 0 - 16
        case  8:                                                //  | ignorieren
        case  9:                                                //  |
        case 10:                                                //  |
        case 11:                                                //  |
        case 12:                                                //  |
        case 13:                                                //  |
        case 14:                                                //  |
        case 15:                                                //  |
        case 16: break;                                         // -+
        case 17: dlsTime = dcfBit; break;                       // MESZ
        case 18: if (dcfBit == dlsTime) bitNo = -1; break;      // MEZ
        case 19: break;                                         // Schaltsekunde
        case 20: if (!dcfBit) bitNo = -1; break;                // Startbit (1)
        case 21: parity = dcfBit; minute = 0;                   // -+
        case 22:                                                //  |
        case 23:                                                //  | Minute
        case 24:                                                //  |
        case 25:                                                //  |
        case 26:                                                //  |
        case 27: if (dcfBit) minute += BCD[bitNo-21]; break;    // -+
        case 28: if (parity) bitNo = -1; break;                 // Parität Minute
        case 29: parity = dcfBit; hour = 0;                     // -+
        case 30:                                                //  |
        case 31:                                                //  | Stunde
        case 32:                                                //  |
        case 33:                                                //  |
        case 34: if (dcfBit) hour += BCD[bitNo-29]; break;      // -+
        case 35: if (parity) bitNo = -1; break;                 // Parität Stunde
        case 36: parity = dcfBit; day  = 0;                     // -+
        case 37:                                                //  | Kalender-
        case 38:                                                //  | tag:
        case 39:                                                //  | 1 .. 31
        case 40:                                                //  |
        case 41: if (dcfBit) day += BCD[bitNo-36]; break;       // -+
        case 42:                                                // -+ Wochentag
        case 43:                                                //  |  ignorieren
        case 44: break;                                         // -+
        case 45: month = 0;                                     // -+
        case 46:                                                //  | Monat:
        case 47:                                                //  |  1 .. 12
        case 48:                                                //  |
        case 49: if (dcfBit) month += BCD[bitNo-45]; break;     // -+
        case 50: year = 0;                                      // -+
        case 51:                                                //  |
        case 52:                                                //  |
        case 53:                                                //  |  Jahr:
        case 54:                                                //  |  0 .. 99
        case 55:                                                //  |
        case 56:                                                //  |
        case 57: if (dcfBit) year += BCD[bitNo-50]; break;      // -+
        case 58: if (parity) bitNo = -1; break;                 // Parität Datum
        default: bitNo = -1;
    }
    if (bitNo == -1) dcfStatus.goodBits = 0;
    else if (dcfStatus.goodBits < 250) dcfStatus.goodBits++;
    bitNo++;
    dcfStatus.bitCount = (UInt08)bitNo;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedClock(void)
{
    // Systemzeit aktualisieren
    if (DATETIME_IsValid(&sysDateTime))
    {   static Timer update = TIMER_MILLISECONDS(1000);
        if (TIMER_RestartIfExpired(&update)) DATETIME_ShiftSecond(&sysDateTime, 1);
    }
    // Zeitgeber abgleichen
    static Timer update = TIMER_SECONDS(CFG_CLOCK_SYNC_INTERVAL*60);
    if (TIMER_RestartIfExpired(&update))
    {   SInt16 dcfElapsed = dcfStatus.elapsed;
        bool dcfGood = dcfElapsed == 0 || dcfElapsed == 1;
        if (radioTimeEnabled && dcfGood)
        {   // RTC mit Systemzeit stellen
            rtClockFailure = !DS1307_Write(&sysDateTime);
        }
        else
        {   // Systemzeit mit RTC stellen
            DateTime tmpDateTime;
            rtClockFailure = !DS1307_Read(&tmpDateTime);
            if (!rtClockFailure) sysDateTime = tmpDateTime;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller      //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den physikali-     //
//  schen Reset auslöst.                                                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
    doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = msecTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = secTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Kopiert <count> Bytes aus dem EEPROM-Speicher ins RAM.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · ramAddress:  Adresse im Ram.                                             //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* eepAddress, void* ramAddress, UInt16 count)
{
    eeprom_read_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Kopiert <count> Bytes aus dem RAM in den EEPROM-Speicher. Da       //
//  ---------------  EEPROM-Speicher lt. Datenblatt nur ca. 100000 Lösch/Schreibzyklen  //
//  pro Zelle verträgt, wird vor für jedes Bytes vor dem Speichervorgang geprüft, ob    //
//  sich der zu schreibende Wert vom Inhalt der EEP-Speicherzelle unterscheidet. Erst   //
//  danach wird nötigenfalls ein Lösch- und Schreibvorgang durchgeführt.                //
//  Hinweis: Das Beschreiben einer EEPROM-Speicherzelle nimmt mit ca. 8 ms verhältnis-  //
//           mäßig viel Zeit in Anspruch.                                               //
//                                                                                      //
//  Eingabe: · ramAddress:  Adresse im Ram.                                             //
//           · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(const void* ramAddress, void* eepAddress, UInt16 count)
{
    eeprom_update_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als CFG_BUS_INTERBYTE_TIMEOUT  //
//  Millisekunden eintritt.                                                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
    bool  idle = false;
    Atomic
    {   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
    }
    return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in <byte> auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()      +--------------+ BOARD_BusTransmit()       //
//     |//////////+------------------------->|//////////////+--------------------+      //
//     |// IDLE //|                COMPLETED |// TRANSMIT //|                    |      //
//     |//////////|<-------------------------+//////////////|<-------------------+      //
//     +----------+                          +------+-------+                           //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +-------------+                            //
//           | BOARD_BusTakeFailure()        |/////////////|                            //
//           +-------------------------------+// FAILURE //|                            //
//                                           |/////////////|                            //
//                                           +-------------+                            //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat von IDLE nach TRANSMIT und beginnt  //
//  mit mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im //
//  Gange (Zustand TRANSMIT), wird bei Aufruf der Funktion das zu übertragende Daten-   //
//  byte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach Aus-     //
//  sendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat    //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sich ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die    //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAILURE bleiben Aufrufe von BOARD_BusTransmit()   //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
    // Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
    bool  queueFull = false;
    do
    {   Atomic
        {   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
        }
    } while (queueFull);
    // Byte senden
    Atomic
    {   if (busStatus.error) return; // Zustand FAILURE: Nichts tun
        if (busStatus.pending)
        {   // Zustand TRANSMIT: Byte für Übertragung aufreihen
            UInt08 tail = busTxQueueHead + busTxQueueLength;
            if (tail >= CARDINALITY(busTxQueue)) tail -= CARDINALITY(busTxQueue);
            busTxQueue[tail] = byte;
            busTxQueueLength++;
        }
        else
        {   // Zustand IDLE: Byte unmittelbar senden
            busStatus.pending = true;
            busStatus.echo    = false;
            busLoopbackByte   = byte;
            UDR0              = byte;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMIT befindet. Das ist solange der Fall,  //
//  wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
    bool  pending = false;
    Atomic
    {   pending = busStatus.pending;
    }
    return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
    bool  error = false;
    Atomic
    {   error = busStatus.error;
        busStatus.error = false;
    }
    return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in <byte> übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Bei einer Unterbrechung der Empfangssequenz kehrt die Funktion mit     //
//  'false' zurück. Umgekehrt gilt das jedoch nicht: Der Rückgabewert 'false' bedeutet  //
//  nicht zwingend, dass auch die Empfangssequenz bereits unterbrochen ist! Dazu siehe  //
//  BOARD_BusIsReceiving().                                                             //
//                                                                                      //
//  Eingabe: · byte: Referenz, die das nächste zur Abholung bereitstehende Datenbyte    //
//                   aufnimmt.                                                          //
//  Ausgabe: 'true', wenn in <byte> ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
    bool  hasData = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        if (!busRxQueue[head].timeout)
        {   // Keine Zeitüberschreitungs-Markierung gesetzt
            if (head != busRxQueueTail)
            {   // Datenbyte vorhanden
                *byte = busRxQueue[head].data;
                head++;
                if (head == CARDINALITY(busRxQueue)) head = 0;
                busRxQueueHead = head;
                hasData = true;
            }
        }
    }
    return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängende Empfangssequenz, wenn  //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als CFG_BUS_INTERBYTE_TIMEOUT Milli-  //
//  sekunden auseinander (timeout), gilt die Empfangssequenz als unterbrochen.          //
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//    Empfangssequenzen: ABC, DE, F...                                                  //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden. Solange BOARD_BusIsReceiving() mit <true> zurückkehrt, //
//  ist die Empfangssequenz noch im Gange.                                              //
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       UInt08 sequence[SEQUENCE_CAPACITY];   // SEQUENCE_CAPACITY: 1 .. 255           //
//                                                                                      //
//       UInt08                                                                         //
//       TakeNextSequence(void)                                                         //
//       {                                                                              //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           UInt08 length = 0;                                                         //
//           if (BOARD_BusIsReceiving())                                                //
//           {   // Empfang einer Sequenz im Gange                                      //
//               UInt08 byte;                                                           //
//               while (BOARD_BusReceive(&byte)) // Datenbytes abholen und aufsammeln   //
//               {   if (index < sizeof(sequence))                                      //
//                   {   sequence[index] = byte;                                        //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {    // Empfang einer Sequenz nicht (mehr) im Gange                        //
//                length = index;                                                       //
//                index  = 0;                                                           //
//           }                                                                          //
//           return length;   // Rückgabewert: Aktuelle Sequenzlänge                    //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
    bool  timeout = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        timeout = busRxQueue[head].timeout;
        if (head != busRxQueueTail) busRxQueue[head].timeout = false;
    }
    return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
    busActivityIndicator = CFG_BUS_ACT_INDICATOR_DURATION;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusActivity(): Der Bus-Aktivitätsindikator zeigt an, ob aktuell mit dem       //
//  -------------------  Gerät kommuniziert wird. Diese Funktion liefert den Zustand    //
//  des Indikators. Nach Abschluss der Kommunikation bleibt der Indikator noch 'CFG_-   //
//  BUS_ACT_INDICATOR_DURATION' Millisekunden aktiviert.                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusActivity(void)
{
    return busActivityIndicator != 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_ReadUTC(): Liest die Systemuhr aus. Die Zeitinformation liegt als Weltzeit    //
//  ---------------  (UTC) vor. Ist die Systemuhr nicht gestellt, ist die in <dateTime> //
//  übergebene Zeit ungültig.                                                           //
//                                                                                      //
//  Eingabe: · dateTime: DateTime-Struktur, die die UTC-Zeit aufnimmt.                  //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_ReadUTC(DateTime* dateTime)
{
    *dateTime = sysDateTime;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetRadioTimeEnabled(): Aktiviert oder deaktiviert den Funkzeitempfang.        //
//  ---------------------------                                                         //
//                                                                                      //
//  Eingabe: · enable: 'true' = Funkzeitempfang ein.                                    //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetRadioTimeEnabled(bool enable)
{
    radioTimeEnabled = enable;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_IsRadioTimeEnabled(): Ist der Funkzeitempfang aktiviert, liefert die Funktion //
//  --------------------------  'true'.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true' = Funkzeitempfang ist aktuell aktiviert.                            //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_IsRadioTimeEnabled(void)
{
    return radioTimeEnabled;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_LastRadioTimeUpdate(): Informiert, wie viele Minuten seit dem letzten,        //
//  ---------------------------  korrekten Funkzeitempfang vergangen sind. Der Zähler   //
//  läuft bis maximal 28800 (= 20 Tage) und bleibt danach auf diesem Wert stehen. Ist   //
//  der Funkempfang aktuell deaktiviert oder hat das System seit dem letzten Start noch //
//  keine Zeitinformation erhalten, liefert die Funktion INVALID zurück.                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Zeit, seit dem letzten Empf.: 0 .. 28800 [min] oder INVALID.               //
//////////////////////////////////////////////////////////////////////////////////////////
SInt16
BOARD_LastRadioTimeUpdate(void)
{
    if (radioTimeEnabled) return dcfStatus.elapsed;
    return INVALID;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_RadioTimeZone(): Liefert die zuletzt empfangene Funkzeitzone. Mögliche Werte  //
//  ---------------------  sind:                                                        //
//                                       1 = UTC+1 (MEZ)                                //
//                                       2 = UTC+2 (MESZ)                               //
//                                    -128 = Unbekannt                                  //
//                                                                                      //
//  Ist der Funkempfang aktuell deaktiviert oder hat das System seit dem letzten Start  //
//  noch keine Zeitinformation erhalten, liefert die Funktion -128 zurück.              //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Die empfangene Zeitzone als Differenz zur UTC oder -128.                   //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
BOARD_RadioTimeZone(void)
{
    return dcfStatus.timeZone;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_RadioTimeGood(): Informiert über den Empfangszustand der Funkzeit. Der        //
//  ---------------------  Empfang gilt als "Gut", wenn mindestens 5 Sekunden lang ver- //
//  wertbare Funksignale eingehen. Ist der Empfang der Funkzeit abgeschaltet, liefert   //
//  die Funktion 'false'.                                                               //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Empfangszustand: 'true' = "gut", false = "schlecht" oder deaktiviert.      //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_RadioTimeGood(void)
{
    return dcfStatus.goodBits > 4;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_RealTimeClockGood(): Informiert über den Zustand der Echtzeituhr.             //
//  -------------------------                                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true' = Echtzeituhr in Ordnung.                                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_RealTimeClockGood(void)
{
    return !rtClockFailure;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_WriteUTC(): Überträgt die in <dateTime> angegebene Zeitinformation in die     //
//  ----------------  Systemuhr. Ist <dateTime> ungültig, wird die Systemuhr ange-      //
//  halten.                                                                             //
//  Hinweis: Das Stellen der Systemuhr ist nur bei abgeschaltetem Funkzeitempfang       //
//           möglich (s.a. BOARD_SetRadioTimeEnabled()).                                //
//                                                                                      //
//  Eingabe: · dateTime: DateTime-Struktur, die die UTC-Zeitinformation für die System- //
//                       uhr bereitstellt.                                              //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_WriteUTC(const DateTime* dateTime)
{
    sysDateTime = *dateTime;
    rtClockFailure = !DS1307_Write(&sysDateTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(LED led, bool on)
{
    switch (led)
    {   case LED_INFO:   ChgPin(INFO_LED, on); break;
        case LED_SYSTEM: ChgPin(SYS_LED , on); break;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Led(LED led)
{
    switch (led)
    {   case LED_INFO:   return (bool)GetPin(INFO_LED);
        case LED_SYSTEM: return (bool)GetPin(SYS_LED);
        default:         return false;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    if (doNext != DO_STARTUP) Reset();
    DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------
    wdt_enable(WDTO_2S);
    // Ports'n'Pins ------------------------------------------------------------------
    ConfPort(B, _L_,    // PB.0: ...     : ---
                _L_,    // PB.1: ...     : ---
                _L_,    // PB.2: SYS_LED : System-Led, H = Ein
                _L_,    // PB.3: [MOSI]
                _L_,    // PB.4: [MISO]
                _L_,    // PB.5: [SCK]
                _T_,    // PB.6: [XTAL1]
                _T_);   // PB.7: [XTAL2]
    ConfPort(C, _L_,    // PC.0: ...     : ---
                _L_,    // PC.1: ...     : ---
                _L_,    // PC.2: ...     : ---
                _L_,    // PC.3: ...     : ---
                _L_,    // PC.4: ...     : ---
                _L_,    // PC.5: ...     : ---
                _T_,    // PC.6: [RESET]
                _T_);   //
    ConfPort(D, _T_,    // PD.0: [RXD]   : Bus
                _T_,    // PD.1: [TXD]   : Bus
                _L_,    // PD.2: ...     : ---
                _L_,    // PD.3  ...     : ---
                _L_,    // PD.4: ...     : ---
                _T_,    // PD.5: DCF_PIN : DCF-Modul Eingangssignal
                _L_,    // PD.6: LED_INFO: Informations-LED, H = Ein
                _L_);   // PD.7: ...
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
    // Timer 2: Clear Timer on Compare Match (CTC) Mode, Intervall 0.9983 ms -------------
    TCCR2A = BV(WGM21);
    TCCR2B = BV(CS21) | BV(CS20);
    OCR2A  = F_CPU/(32*1000UL)-1;   // 32 = Vorteilung, 1000 = Wiederholrate bei 3.686400
    TIMSK2 = BV(OCIE2A);
    // USART 19200 Baud, 8 Data, 1 Stop, No Parity, Async. -------------------------------
    UBRR0  = F_CPU/16L/19200-1;
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(TXCIE0) | BV(RXEN0) | BV(TXEN0);
    UCSR0C = BV(UCSZ01) | BV(UCSZ00);
    // Sonstiges -------------------------------------------------------------------------
    EnableInterrupts();
    doNext = DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    wdt_reset();
    ProceedDcfDecoding();
    ProceedClock();
    // 1ms Zeitbarriere
    UInt16 ticks = BOARD_MSecTicks();
    {   static UInt08 ms = 0;
        if (ms == (UInt08)ticks) return;
        ms = (UInt08)ticks;
    }
    // Bus-Aktivitätsindikator
    if (busActivityIndicator > 0)
    {   busActivityIndicator--;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    while (BOARD_BusIsTransmitting());
    switch (doNext)
    {   case DO_ENTER_PGMLDR: EnterPgmLdr();        break;
        case DO_STARTUP:      Reset();              break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
